{*******************************************************}
{                                                       }
{       Advanced SQL statement parser                   }
{       Copyright (c) 2001 AS Gaiasoft                  }
{       Created by Gert Kello                           }
{                                                       }
{*******************************************************}

unit gaSelectStm;

interface

uses
  gaAdvancedSQLParser, gaSQLParserHelperClasses;

type
  TSelectStatementState = (sssNone, sssFieldList, sssFromList, sssJoinClause,
    sssWhereClause, sssGroupBy, sssHavingClause, sssOrderBy, sssUnion,
    sssStatementComplete);

  TSelectStatementType = (sstSelectAll, sstSelectDistinct);

  TSelectStatementStates = set of TSelectStatementState;

  TgaSelectSQLStatement = class (TgaCustomSQLStatement)
  private
    FOrderByClause: TgaSQLOrderByList;
    FSelectType: TSelectStatementType;
    FStatementFields: TgaSQLFieldList;
    FStatementState: TSelectStatementState;
    FStatementTables: TgaSQLTableList;
    FWhereClause: TgaSQLWhereExpression;
    procedure SetStatementState(const Value: TSelectStatementState);
  protected
    function GetNewStatementState: TSelectStatementState;
    function GetStatementType: TSQLStatementType; override;
    procedure ModifyStatementInNormalState(Sender: TObject; AToken: 
            TgaSQLTokenObj); override;
    procedure ParseFieldList(AToken: TgaSQLTokenObj);
    procedure ParseGroupBy;
    procedure ParseHavingClause;
    procedure ParseJoin;
    property StatementState: TSelectStatementState read FStatementState write 
            SetStatementState;
  public
    constructor Create(AOwner: TgaAdvancedSQLParser); override;
    destructor Destroy; override;
    procedure ParseOrderBy(AToken: TgaSQLTokenObj);
    property OrderByClause: TgaSQLOrderByList read FOrderByClause;
    property SelectType: TSelectStatementType read FSelectType;
    property StatementFields: TgaSQLFieldList read FStatementFields;
    property StatementTables: TgaSQLTableList read FStatementTables;
    property WhereClause: TgaSQLWhereExpression read FWhereClause;
  end;
  
const
  SelectAllowedNextState: array[TSelectStatementState] of TSelectStatementStates =
    ( {sssNone} [sssFieldList],
      {sssFieldList}[sssFromList],
      {sssFromList} [sssJoinClause, sssWhereClause, sssGroupBy, sssOrderBy,
                      sssUnion, sssStatementComplete],
      {sssJoinClause} [sssWhereClause, sssGroupBy, sssOrderBy, sssUnion,
                        sssStatementComplete],
      {sssWhereClause} [sssGroupBy, sssOrderBy, sssUnion, sssStatementComplete],
      {sssGroupBy} [sssHavingClause, sssOrderBy, sssUnion, sssStatementComplete],
      {sssHavingClause} [sssOrderBy, sssUnion, sssStatementComplete],
      {sssOrderBy} [sssUnion, sssStatementComplete],
      {sssUnion} [], {unions will be parsed by another class}
      {sssStatementComplete} []);

implementation

uses
  SysUtils, TypInfo, gaBasicSQLParser, gaSQLParserConsts;

{
**************************** TgaSelectSQLStatement *****************************
}
constructor TgaSelectSQLStatement.Create(AOwner: TgaAdvancedSQLParser);
begin
  inherited Create(AOwner);
  FStatementState := sssNone;
  FStatementFields := TgaSQLFieldList.Create(Self);
  FStatementTables := TgaSQLTableList.Create(Self);
end;

destructor TgaSelectSQLStatement.Destroy;
begin
  FOrderByClause.Free;
  FStatementFields.Free;
  FStatementTables.Free;
  FWhereClause.Free;
  inherited Destroy;
end;

function TgaSelectSQLStatement.GetNewStatementState: TSelectStatementState;
var
  TokenStr: string;
begin
  Result := StatementState;
  TokenStr := UpperCase(OwnerParser.TokenString);
  if OwnerParser.TokenType = stSymbol then
  begin
    if TokenStr = 'SELECT' then
      Result := sssFieldList
    else if TokenStr = 'FROM' then
      Result := sssFromList
    else if TokenStr = 'WHERE' then
      Result := sssWhereClause
    else if TokenStr = 'GROUP' then
      Result := sssGroupBy
    else if TokenStr = 'HAVING' then
      Result := sssHavingClause
    else if TokenStr = 'ORDER' then
      Result := sssOrderBy
    else if TokenStr = 'UNION' then
      Result := sssUnion
    else if (TokenStr = 'INNER') or (TokenStr = 'JOIN') or
      (TokenStr = 'LEFT') or (TokenStr = 'RIGHT') or (TokenStr = 'FULL') then
      Result := sssJoinClause;
  end else if OwnerParser.TokenType = stEnd then
    Result := sssStatementComplete;
end;

function TgaSelectSQLStatement.GetStatementType: TSQLStatementType;
begin
  Result := sstSelect;
end;

procedure TgaSelectSQLStatement.ModifyStatementInNormalState(Sender: TObject; 
        AToken: TgaSQLTokenObj);
begin
  inherited;
  StatementState := GetNewStatementState;
  if StatusCode <> 0 then
    raise Exception.CreateFmt('Statement status code is %d', [StatusCode]);
  if InternalStatementState > 0 then
    case StatementState of
      sssNone:
        {the statement starts with comment or whitespace};
      sssFieldList:
        ParseFieldList(AToken);
      sssFromList:
        StatementTables.AddToken(AToken);
      sssWhereClause: begin
        if InternalStatementState = 1 then
        begin
          FWhereClause := TgaSQLWhereExpression.Create(Self);
          InternalStatementState := 2;
        end;
        WhereClause.AddToken(AToken);
      end;
      sssGroupBy:
        ParseGroupBy;
      sssHavingClause:
        ParseHavingClause;
      sssJoinClause:
        ParseJoin;
      sssOrderBy:
        ParseOrderBy(AToken);
      sssStatementComplete:
        DoStatementComplete;
      else
        raise Exception.CreateFmt(SUnknownStatementState,
          [ClassName, GetEnumName(TypeInfo(TSelectStatementState), Ord(StatementState))]);
    end
  else
    InternalStatementState := 1;
end;

procedure TgaSelectSQLStatement.ParseFieldList(AToken: TgaSQLTokenObj);
begin
  if InternalStatementState = 1 then
  begin
    if AToken.TokenType in [stDelimitier, stComment] then
      // continue
    else begin
      if AToken.TokenSymbolIs('ALL') then
        FSelectType := sstSelectAll
      else if AToken.TokenSymbolIs('DISTINCT') then
        FSelectType := sstSelectDistinct
      else
      StatementFields.AddToken(AToken);
      InternalStatementState := 2;
    end;
  end else
    StatementFields.AddToken(AToken);
end;

procedure TgaSelectSQLStatement.ParseGroupBy;
begin
  // #todo2
end;

procedure TgaSelectSQLStatement.ParseHavingClause;
begin
  // #todo2
end;

procedure TgaSelectSQLStatement.ParseJoin;
begin
  // #todo2
end;

procedure TgaSelectSQLStatement.ParseOrderBy(AToken: TgaSQLTokenObj);
begin
  case InternalStatementState of
    1: begin
      case AToken.TokenType of
        stDelimitier, stComment:
          ;// Do nothing, wait for "by"
        stSymbol:
          if AToken.TokenSymbolIs('BY') then
          begin
            InternalStatementState := 2;
            FOrderByClause := TgaSQLOrderByList.Create(Self);
          end else
            StatusCode := errWrongKeywordSequence;
      else
        StatusCode := errWrongKeywordSequence;
      end;
    end;
    else
      OrderByClause.AddToken(AToken);
  end;
end;

procedure TgaSelectSQLStatement.SetStatementState(const Value: 
        TSelectStatementState);
begin
  if StatementState <> Value then
  begin
    if StatusCode = 0 then
      if Value in SelectAllowedNextState[FStatementState] then
      begin
        DoBeforeStatementStateChange(Ord(Value));
        FStatementState := Value;
        InternalStatementState := 0;
        DoAfterStatementStateChange;
      end else
        StatusCode := errWrongKeywordSequence;
    if Value = sssStatementComplete then
    begin
      DoBeforeStatementStateChange(Ord(Value));
      FStatementState := Value;
      InternalStatementState := 0;
      DoAfterStatementStateChange;
    end;
  end;
end;

end.
